// SPDX-License-Identifier: BUSL-1.1
pragma solidity 0.8.21;

import "@openzeppelin/contracts-upgradeable/access/OwnableUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/proxy/ClonesUpgradeable.sol";
import "../abstract/BlastPointsReceiver.sol";
import "./StakingPool.sol";

/**
 * @title StakingAggregator
 * @notice Contract for creating staking pool instances (various periods + APY).
 */
contract StakingAggregator is OwnableUpgradeable {
    address public baseImplementation;

    address[] public pools;

    event PoolCreated(
        address indexed pool,
        uint256 indexed period,
        uint256 indexed apy,
        address token,
        uint256 limit
    );

    /// @dev Creates a new contract
    constructor() {
        _disableInitializers();
    }

    /// @notice Inits a new contract
    /// @param baseImplementation_ staking pool base implementation address
    /// @param admin_ aggregator admin address
    function init(address baseImplementation_, address admin_) external initializer {
        require(baseImplementation_ != address(0), 'Invalid base implementation');
        require(admin_ != address(0), 'Invalid admin');

        baseImplementation = baseImplementation_;
        _transferOwnership(admin_);
    }

    /// @notice Creates a new staking pool
    /// @param stakingToken_ address of the token to be staked
    /// @param stakingPeriod_ pool staking period in seconds
    /// @param apy_ pool apy
    /// @param stakingTokenLimit_ pool total supply limitter
    function createStakingPool(
        address stakingToken_,
        address owner_,
        uint256 stakingPeriod_,
        uint256 apy_,
        uint256 stakingTokenLimit_,
        IBlastPoints blastPoints_,
        address pointsOperator_
    ) external onlyOwner {
        address pool = ClonesUpgradeable.clone(baseImplementation);
        StakingPool(pool).init(
            stakingToken_,
            owner_,
            stakingPeriod_,
            apy_,
            stakingTokenLimit_,
            blastPoints_,
            pointsOperator_
        );
        pools.push(pool);

        emit PoolCreated(
            pool,
            stakingPeriod_,
            apy_,
            stakingToken_,
            stakingTokenLimit_
        );
    }

    /// @dev Allows to get a slice of pools array
    /// @param offset Starting index in pools array
    /// @param length return array length
    /// @return Array-slice of pools
    function getPoolsSlice(
        uint256 offset,
        uint256 length
    ) external view returns (address[] memory) {
        require(length != 0, 'Zero length');
        require(offset + length <= pools.length, 'Invalid offset + length');

        address[] memory res = new address[](length);

        for (uint256 i; i < length; ) {
            res[i] = pools[i + offset];
            unchecked {
                ++i;
            }
        }

        return res;
    }

    /// @dev Allows to view pools array length
    /// @return length of user pools array
    function getPoolsLength() external view returns (uint256) {
        return pools.length;
    }
}
