// SPDX-License-Identifier: BUSL-1.1
pragma solidity 0.8.21;

import "@openzeppelin/contracts-upgradeable/access/OwnableUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/security/PausableUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/token/ERC20/utils/SafeERC20Upgradeable.sol";
import "@openzeppelin/contracts-upgradeable/utils/structs/EnumerableSetUpgradeable.sol";
import "../abstract/BlastPointsReceiver.sol";

/**
 * @title StakingFlexible
 * @notice Contract for staking tokens in order to earn rewards.
 * Any user can make multiple stakes. Reward earn period is practically unlimited.
 */

contract StakingFlexible is OwnableUpgradeable, PausableUpgradeable {
    using SafeERC20Upgradeable for IERC20Upgradeable;
    using EnumerableSetUpgradeable for EnumerableSetUpgradeable.UintSet;
    using EnumerableSetUpgradeable for EnumerableSetUpgradeable.AddressSet;

    // TYPES

    struct Stake {
        uint256 id;
        uint256 amount;
        uint256 earned;
        uint256 claimed;
        uint256 userRewardPerTokenPaid;
        address owner;
        uint32 timestamp;
        uint32 unstakedAtBlockNumber;
        uint32 unstakedAtBlockTimestamp;
    }

    // CONSTANTS

    uint256 public constant ACCURACY = 1e18;
    uint256 public constant TREASURY_STAKE_AMOUNT = 1; // 1 wei (to simplify calculations)
    uint256 public constant MIN_REWARDS_DURATION = 1 minutes; // 1 min
    uint256 public constant MAX_STAKE = 10 ** 70;
    // reward rate will not fall into overflow
    // even if reward is being added during 1000 min possible reward periods in a row
    uint256 public constant MAX_RATE = type(uint128).max / (1000 * MIN_REWARDS_DURATION);
    uint256 public constant MAX_REWARDS_DURATION = 10 * 365 days; // 10 years (to prevent overflow)
    uint256 public constant PRINCIPAL_FEE_LIMITTER = 10_00;
    uint256 public constant ONE_HUNDRED = 100_00;

    // STATE VARIABLES

    IERC20Upgradeable public stakingToken;
    IERC20Upgradeable public rewardToken;

    address public feeWallet;

    uint128 public rewardRate;
    uint128 public rewardsRemaining;

    uint256 public rewardPerTokenStored;

    uint64 public periodFinish;

    // rewardsDuration is not  always equal to periodFinish - periodStart
    uint64 public rewardsDuration;

    uint64 public lastUpdateTime;

    // variable for correct reward calculations in case  of rewardsDuration changes
    uint64 public periodStart;

    // sum of all rewards distributed to users in all periods
    uint256 public cumulativeDistributedReward;
    // sum of claimed reawards in one current reward period
    uint256 public currentRewardPaid;
    // sum of claimed reawards in all reward periods
    uint256 public cumulativeRewardPaid;

    // sum of unused rewards by all users in all owner locks in the current period
    // unused reward is set = 0 in each notify
    uint256 public currentPeriodUnusedRewardsSum;

    uint256 public unusedRewardsExtracted;

    mapping(uint256 => Stake) public stakes;
    mapping(address => uint256[]) public userInactiveStakes;
    mapping(address => EnumerableSetUpgradeable.UintSet) private idsByUser;

    mapping(address => uint256) public userBalances;

    uint256 public globalId;

    // hide real total supply
    uint256 private totalSupplyInternal;
    // number of currenlty active stakes
    uint128 private numOfActiveStakesInternal;

    // first reward period creation timestamp
    uint128 public firstRewardPeriodStart;

    uint256 public unstakePrincipalFee;

    // active stake holder
    EnumerableSetUpgradeable.AddressSet private holders;

    bool public finished;

    // EVENTS

    event RewardAdded(uint256 reward);
    event Staked(
        address indexed user,
        uint256 indexed id,
        uint256 amount,
        uint256 totalSupply
    );
    event Withdrawn(
        address indexed user,
        uint256 indexed id,
        uint256 amount,
        uint256 totalSupply
    );
    event RewardPaid(address indexed user, uint256 indexed id, uint256 reward);
    event RewardsDurationUpdated(uint256 indexed newDuration);
    event Finished(address indexed recepient);
    event WithdrawnTokens(address indexed recepient, uint256 indexed amount);
    event SetUnstakeFee(uint256 indexed fee);
    event SetFeeWallet(address indexed feeWallet);

    // blast points events

    event Deposit(address indexed user, uint256 amount);
    event Claimed(address indexed user, uint256 reward);
    event Withdraw(address indexed user, uint256 amount, uint256 reward);

    //  FUNCTIONS

    constructor() {
        _disableInitializers();
    }

    // external

    /// @notice Inits a new contract
    /// @param stakingToken_ staking token address
    /// @param rewardToken_ reward token address
    /// @param owner_ owner address
    function init(
        IERC20Upgradeable stakingToken_,
        IERC20Upgradeable rewardToken_,
        address owner_,
        uint256 unstakePrincipalFee_,
        address feeWallet_,
        IBlastPoints blastPoints_,
        address pointsOperator_
    ) external initializer {
        require(address(stakingToken_) != address(0), 'Invalid address');
        require(address(rewardToken_) != address(0), 'Invalid address');
        require(owner_ != address(0), 'Invalid address');
        require(unstakePrincipalFee_ <= PRINCIPAL_FEE_LIMITTER, 'Fee too high');
        require(feeWallet_ != address(0), 'Invalid fee wallet');

        if (address(blastPoints_) != address(0)) {
            if (pointsOperator_ == address(0)) revert('Invalid blast points');
            blastPoints_.configurePointsOperator(pointsOperator_);
        } else if (pointsOperator_ != address(0)) revert('Invalid operator');

        stakingToken = stakingToken_;
        rewardToken = rewardToken_;
        rewardsDuration = 365 days;
        unstakePrincipalFee = unstakePrincipalFee_;
        feeWallet = feeWallet_;

        _transferOwnership(owner_);
    }

    /// @dev Allows user to stake tokens
    /// @param amount of token to stake
    function stake(uint256 amount) external whenNotPaused {
        require(amount != 0, 'Cannot stake 0');
        require(amount <= MAX_STAKE, 'Stake amount too high');
        require(periodFinish > block.timestamp, 'Reward period not activated');

        // if no active stakers were previously => unstake owner stake to stop accumulating unused rewards
        if (totalSupply() == 0) {
            // change stakes[0].earned
            _updateReward(0, true);
            (, uint256 unusedRewards) = _withdrawInternal(0, address(this));
            currentPeriodUnusedRewardsSum += unusedRewards;
        } else {
            // do not change stakes[0]
            _updateReward(0, false);
        }

        // execute stake
        _stakeInternal(amount, msg.sender);

        emit Staked(msg.sender, globalId, amount, totalSupply());
        emit Deposit(msg.sender, amount);

        stakingToken.safeTransferFrom(msg.sender, address(this), amount);
    }

    /// @dev Allows user to withdraw staked tokens + claim earned rewards
    /// @param id Stake id
    function withdraw(uint256 id) external {
        _updateReward(id, true);
        // CHANGE STATE & EXECUTE ALL CHECKS
        (uint256 amount, uint256 reward) = _withdrawInternal(id, msg.sender);

        // if no active stakers left => create a owner stake to accumulate unused rewards
        if (totalSupply() == 0) {
            _stakeInternal(TREASURY_STAKE_AMOUNT, address(this));
        }

        emit Withdrawn(msg.sender, id, amount, totalSupply());

        // ALL TOKENS TRANSFERS -------------------------------------------------------

        // REWARDS

        emit RewardPaid(msg.sender, id, reward);
        emit Withdraw(msg.sender, amount, reward);

        if (reward != 0) {
            rewardToken.safeTransfer(msg.sender, reward);
        }

        // PRINCIPAL

        uint256 fee = (amount * unstakePrincipalFee) / ONE_HUNDRED;
        stakingToken.safeTransfer(msg.sender, amount - fee);
        if (fee != 0) stakingToken.safeTransfer(feeWallet, fee);
    }

    /// @dev Allows user to claim earned rewards without unstaking the position
    /// @param id Stake id
    function claim(uint256 id) external {
        _updateReward(id, true);
        // CHANGE STATE & EXECUTE ALL CHECKS
        uint256 reward = _claimInternal(id, msg.sender);

        require(reward != 0, 'Nothing to claim');

        // ALL TOKENS TRANSFERS -------------------------------------------------------

        // REWARDS

        emit RewardPaid(msg.sender, id, reward);
        emit Claimed(msg.sender, reward);

        // reward != 0 due to technical design
        rewardToken.safeTransfer(msg.sender, reward);
    }

    /// @dev Allows owner to set new reward rate
    /// @param newRewardRate is amount of token weis will be distributed per second (rewards / rewardsDuration)
    function notifyRewardAmount(uint128 newRewardRate) external onlyOwner {
        _updateReward(0, false);
        _notifyRewardAmount(newRewardRate);
    }

    /// @dev Allows owner to set previous reward rate
    function notifyRewardAmountPrevious() external onlyOwner {
        _updateReward(0, false);
        uint128 rate = rewardRate;
        require(rate != 0, 'Previous reward rate is too low');
        _notifyRewardAmount(rate);
    }

    /// @dev Allows owner to stop staking and extract extra liquidity
    /// @param recepient is the extra liquidity recepient
    function finish(address recepient) external onlyOwner {
        require(!finished, 'Finished');
        require(recepient != address(0), 'Zero recepient');
        uint256 periodFinishLocal = periodFinish;
        require(
            block.timestamp > periodFinishLocal && periodFinishLocal != 0,
            'Only after period end'
        );

        // lastTimeRewardApplicable == periodFinish  (period is already over)
        uint256 previousPeriodRewardDuration = periodFinishLocal - periodStart;

        // no need to unstake owner stake (no active reward period + all rewards are counted in getUnusedRewards())

        // update cumulative variables
        // use REAL previous period rewards duration here
        uint256 currentRewardDistributed = rewardRate * previousPeriodRewardDuration;

        cumulativeDistributedReward += currentRewardDistributed;
        cumulativeRewardPaid += currentRewardPaid;
        currentRewardPaid = 0;

        uint256 unusedReward = getUnusedRewards();

        uint256 unclaimedReward = cumulativeDistributedReward -
            cumulativeRewardPaid -
            unusedReward -
            unusedRewardsExtracted;

        IERC20Upgradeable rewardTokenLocal = rewardToken;
        IERC20Upgradeable stakingTokenLocal = stakingToken;

        uint256 rewardTokenBalance = rewardTokenLocal.balanceOf(address(this));
        uint256 stakingTokenBalance = stakingTokenLocal.balanceOf(address(this));

        finished = true;

        // unusedRewards - from contract
        // unclaimed rewards - on contract

        emit Finished(recepient);

        if (rewardTokenLocal == stakingTokenLocal) {
            uint256 supply = totalSupply();

            if (rewardTokenBalance - (supply + unclaimedReward) != 0) {
                rewardTokenLocal.safeTransfer(
                    recepient,
                    rewardTokenBalance - (supply + unclaimedReward)
                );
            }
        } else {
            if (rewardTokenBalance - unclaimedReward != 0) {
                rewardTokenLocal.safeTransfer(
                    recepient,
                    rewardTokenBalance - unclaimedReward
                );
            }

            uint256 supply = totalSupply();
            if (stakingTokenBalance - supply != 0) {
                stakingTokenLocal.safeTransfer(recepient, stakingTokenBalance - supply);
            }
        }
    }

    /// @dev Allows owner to stop staking and extract extra liquidity
    /// @param recepient is the extra liquidity recepient
    function withdrawOtherTokens(address token, address recepient) external onlyOwner {
        require(token != address(0), 'Invalid token');
        require(recepient != address(0), 'Zero recepient');
        require(token != address(stakingToken), 'Token == staking token');
        require(token != address(rewardToken), 'Token == reward token');

        uint256 balance = IERC20Upgradeable(token).balanceOf(address(this));
        require(balance != 0, 'Zero balance');

        emit WithdrawnTokens(recepient, balance);
        IERC20Upgradeable(token).safeTransfer(recepient, balance);
    }

    /// @dev Allows owner to set period of rewards distribution
    /// @param rewardsDuration_ is new rewardsDuration amount
    function setRewardsDuration(uint64 rewardsDuration_) external onlyOwner {
        require(block.timestamp > periodFinish, 'Previous period not complete');
        require(
            rewardsDuration_ >= MIN_REWARDS_DURATION,
            'Rewards duration is too short'
        );
        require(rewardsDuration_ <= MAX_REWARDS_DURATION, 'Rewards duration is too long');
        require(rewardsDuration_ != rewardsDuration, 'Rewards duration duplicate');
        rewardsDuration = rewardsDuration_;
        emit RewardsDurationUpdated(rewardsDuration_);
    }

    /// @dev Sets paused state for the contract (can be called by the owner only)
    /// @param state paused flag
    function setPaused(bool state) external onlyOwner {
        if (state) {
            _pause();
        } else {
            _unpause();
        }
    }

    /// @dev Allows owner to set the new unstake principal fee
    /// @param unstakePrincipalFee_ is new unstake principal fee
    function setUnstakeFee(uint256 unstakePrincipalFee_) external onlyOwner {
        require(unstakePrincipalFee_ <= PRINCIPAL_FEE_LIMITTER, 'Fee too high');
        require(unstakePrincipalFee_ != unstakePrincipalFee, 'Duplicate');

        unstakePrincipalFee = unstakePrincipalFee_;

        emit SetUnstakeFee(unstakePrincipalFee_);
    }

    /// @dev Allows owner to set the new fee wallet
    /// @param feeWallet_ is new unstake principal fee
    function setFeeWallet(address feeWallet_) external onlyOwner {
        require(feeWallet_ != address(0), 'Invalid fee wallet');
        require(feeWallet_ != feeWallet, 'Duplicate');

        feeWallet = feeWallet_;

        emit SetFeeWallet(feeWallet_);
    }

    // external view

    /// @dev Allows to view staking amount of selected user
    /// @param account to view balance
    /// @return balance of staking tokens for selected account
    function balanceOf(address account) external view returns (uint256) {
        return userBalances[account];
    }

    /// @dev Allows to view user`s stake ids quantity
    /// @param user user account
    /// @return length of user ids array
    function getUserStakeIdsLength(address user) external view returns (uint256) {
        return idsByUser[user].length();
    }

    /// @dev Allows to view if a user has a stake with specific id
    /// @param user user account
    /// @param id stake id
    /// @return bool flag (true if a user has owns the id)
    function hasStakeId(address user, uint256 id) external view returns (bool) {
        return idsByUser[user].contains(id);
    }

    /// @dev Allows to get a slice user stakes array
    /// @param user user account
    /// @param startIndex Starting index in user ids array
    /// @param length return array length
    /// @return Array-slice of user stakes
    function getUserStakesSlice(
        address user,
        uint256 startIndex,
        uint256 length
    ) external view returns (Stake[] memory) {
        require(length != 0, 'Zero length');
        require(
            startIndex + length <= idsByUser[user].length(),
            'Invalid startIndex + length'
        );
        Stake[] memory userStakes = new Stake[](length);
        for (uint256 i; i < length; ) {
            uint256 stakeId = idsByUser[user].at(i + startIndex);
            userStakes[i] = stakes[stakeId];
            userStakes[i].earned = earned(stakeId);

            unchecked {
                ++i;
            }
        }

        return userStakes;
    }

    /// @dev Allows to get a slice user stakes history array
    /// @param user user account
    /// @param startIndex Starting index in user ids array
    /// @param length return array length
    /// @return Array-slice of user stakes history
    function getUserInactiveStakesSlice(
        address user,
        uint256 startIndex,
        uint256 length
    ) external view returns (Stake[] memory) {
        require(length != 0, 'Zero length');
        require(
            startIndex + length <= userInactiveStakes[user].length,
            'Invalid startIndex + length'
        );
        Stake[] memory userStakes = new Stake[](length);

        for (uint256 i; i < length; ) {
            uint256 stakeId = userInactiveStakes[user][i + startIndex];
            userStakes[i] = stakes[stakeId];

            unchecked {
                ++i;
            }
        }
        return userStakes;
    }

    /// @dev Allows to view user`s closed stakes quantity
    /// @param user user account
    /// @return length of user closed stakes array
    function getUserInactiveStakesLength(address user) external view returns (uint256) {
        return userInactiveStakes[user].length;
    }

    /// @dev Allows to view total cumulative unclaimed rewards (earned by all users)
    /// @return rewards (nominated in token weis)
    function getTotalRewards() external view returns (uint256) {
        if (totalSupply() != 0) {
            uint256 timePassed = lastTimeRewardApplicable() - periodStart;

            uint256 totalDistributedRewards = cumulativeDistributedReward;

            if (!finished) {
                totalDistributedRewards += rewardRate * timePassed;
            }

            uint256 totalClaimedRewards = cumulativeRewardPaid + currentRewardPaid;

            return
                totalDistributedRewards -
                totalClaimedRewards -
                getUnusedRewards() -
                unusedRewardsExtracted;
        } else {
            return 0;
        }
    }

    /// @dev Returns the approximate APR for a specific stake position not including penalties
    /// @return APR value multiplied by 10**18
    function getPotentialAPR() external view returns (uint256) {
        uint256 _tvl = totalSupply();

        if (_tvl == 0) return 0;
        // earned user value in tokens for one second including share (1e18 for accuracy)
        return (uint256(rewardRate) * ACCURACY * 365 days * 100) / (_tvl);
    }

    /// @dev Allows to view the number of stake holders
    /// @return length of stake holders array
    function getHoldersLength() external view returns (uint256) {
        return holders.length();
    }

    /// @dev Allows to get a slice of stake holders array
    /// @param startIndex Starting index in stake holders array
    /// @param length return array length
    /// @return Array-slice of stake holders
    function getHoldersSlice(
        uint256 startIndex,
        uint256 length
    ) external view returns (address[] memory) {
        require(length != 0, 'Zero length');
        require(startIndex + length <= holders.length(), 'Invalid startIndex + length');
        address[] memory holdersArray = new address[](length);
        for (uint256 i = 0; i < length; ) {
            holdersArray[i] = holders.at(i + startIndex);
            unchecked {
                ++i;
            }
        }
        return holdersArray;
    }

    /// @notice Returns the number of currently active stakes
    /// @dev Excludes the number of owner`s active stakes
    /// @return numOfActiveStakesInternal - number of active stakes, made by users
    function numOfActiveStakes() external view returns (uint256) {
        return numOfActiveStakesInternal - idsByUser[address(this)].length();
    }

    // public

    /// @notice Returns the user`s stakes total supply
    /// @dev Excludes reward owner stake supply
    /// @return totalSupplyInternal - total amount of staked tokens (in weis)
    function totalSupply() public view returns (uint256) {
        return totalSupplyInternal - userBalances[address(this)];
    }

    /// @dev Allows to view last action time in the contract
    /// @return timestamp of last action
    function lastTimeRewardApplicable() public view returns (uint64) {
        return block.timestamp < periodFinish ? uint64(block.timestamp) : periodFinish;
    }

    /// @dev Allows to view already distributed rewards
    /// @return rewardPerTokenStored for current moment
    function rewardPerToken() public view returns (uint256) {
        uint256 totalSupplyInternalLocal = totalSupplyInternal;
        // can be == 0 before the first owner stake in made
        if (totalSupplyInternalLocal == 0) {
            return rewardPerTokenStored;
        }
        // can not overflow here, rewardRate * duration * ACCURACY <= uint128_max * 10 ** 18 << uint256 max
        return
            rewardPerTokenStored +
            (((lastTimeRewardApplicable() - lastUpdateTime) *
                uint256(rewardRate) *
                ACCURACY) / totalSupplyInternalLocal);
    }

    /// @dev Allows to view current user earned rewards
    /// @param id to view rewards
    /// @return earnedAmount of rewards for selected user
    function earned(uint256 id) public view returns (uint256) {
        Stake storage _stake = stakes[id];
        if (_stake.unstakedAtBlockNumber == 0) {
            // ACTIVE STAKE => calculate amount + increase reward per token
            // can not overflow here, amount * (rewardRate * duration * ACCURACY / total supply)
            return
                uint256(
                    (_stake.amount * (rewardPerToken() - _stake.userRewardPerTokenPaid)) /
                        ACCURACY +
                        _stake.earned
                );
        }

        // INACTIVE STAKE
        return 0;
    }

    /// @notice Returns the amount of unused reward tokens
    /// @return reward tokens amount in weis
    function getUnusedRewards() public view returns (uint256) {
        // makes return amount always updated real time
        return earned(0) + currentPeriodUnusedRewardsSum;
    }

    // private

    /// @dev Allows to set reward rate
    /// @param newRewardRate is amount of token weis will be distributed per second (rewards / rewardsDuration)
    function _notifyRewardAmount(uint128 newRewardRate) private {
        require(!finished, 'Finished');
        require(newRewardRate <= MAX_RATE, 'Reward rate too high');

        // reward for the incoming period => use the new rewardsDuration
        uint64 rewardsDurationLocal = rewardsDuration;
        uint128 rewardRateOld = rewardRate;
        uint256 reward = newRewardRate * rewardsDurationLocal;
        uint256 periodFinishOld = periodFinish;

        bool restakeOwnerPosition = true;
        if (firstRewardPeriodStart == 0) {
            // open owner stake in the first reward period (stake 1 wei to to accumulate unused rewards)
            _stakeInternal(TREASURY_STAKE_AMOUNT, address(this));

            firstRewardPeriodStart = uint128(block.timestamp);
            restakeOwnerPosition = false;
        }

        // update cumulative variables
        // add rewards from teh previous period => use previous reward period start
        uint256 currentRewardDistributed = periodFinishOld == 0
            ? 0
            : rewardRateOld * (lastTimeRewardApplicable() - periodStart);

        cumulativeDistributedReward += currentRewardDistributed;
        cumulativeRewardPaid += currentRewardPaid;
        currentRewardPaid = 0;

        // update rate with the reward from the incoming period => use the updated rewards duration
        if (block.timestamp >= periodFinishOld) {
            // can not overflow here (MAX Numerator = Const * uint128_max * uint64),
            // Const < 100 <= num of reward periods, min period = 1 year
            rewardRate = uint128((reward + rewardsRemaining) / rewardsDurationLocal);
            // can not overfow here => MAX value = uint64_max - 1
            rewardsRemaining = uint128(
                (reward + rewardsRemaining) % rewardsDurationLocal
            );
        } else {
            uint64 remaining = uint64(periodFinishOld) - uint64(block.timestamp);
            uint256 leftover = remaining * rewardRateOld + rewardsRemaining;
            // can not overflow here (MAX Numerator = Const * uint128_max * uint64),
            // Const < 100 <= num of reward periods, min period = 1 year
            rewardRate = uint128((reward + leftover) / rewardsDurationLocal);
            // can not overfow here => MAX value = uint64_max - 1
            rewardsRemaining = uint128((reward + leftover) % rewardsDurationLocal);
        }

        uint256 rewardRateNew = rewardRate;

        require(rewardRateNew != 0, 'Actual reward rate is too low');

        // Ensure the provided reward amount is not more than the balance in the contract.
        // This keeps the reward rate in the right range, preventing overflows due to
        // very high values of rewardRate in the earned and rewardsPerToken functions;
        uint256 balance = address(rewardToken) == address(stakingToken)
            ? rewardToken.balanceOf(address(this)) - totalSupply()
            : rewardToken.balanceOf(address(this));

        bool updateUnusedReward = (totalSupply() == 0);
        uint256 unusedReward = getUnusedRewards();

        uint256 unclaimedReward = cumulativeDistributedReward -
            cumulativeRewardPaid -
            unusedReward -
            unusedRewardsExtracted;

        // unusedRewards are transferred to the owner in each notify
        currentPeriodUnusedRewardsSum = 0;
        // increase total unused rewards extracted after each period
        unusedRewardsExtracted += unusedReward;

        // if no active stakers were previously => unstake owner stake to stop accumulating unused rewards
        if (updateUnusedReward && restakeOwnerPosition) {
            _updateReward(0, true);
            // no update for cumulative unused reward here (already called getUnusedRewards() before, using earned())
            _withdrawInternal(0, address(this));
        }

        // Ensure the balance in the contract is more than the maximum reward + unclaimedRewards
        require(
            rewardRateNew * rewardsDurationLocal + unclaimedReward + unusedReward <=
                balance,
            'Provided reward too high'
        );

        lastUpdateTime = uint64(block.timestamp);
        periodFinish = uint64(block.timestamp) + rewardsDurationLocal;
        periodStart = uint64(block.timestamp);

        if (updateUnusedReward && restakeOwnerPosition) {
            _stakeInternal(TREASURY_STAKE_AMOUNT, address(this));
        }

        if (unusedReward != 0)
            IERC20Upgradeable(rewardToken).safeTransfer(owner(), unusedReward);

        emit RewardAdded(reward);
    }

    /// @dev Internal function for staking tokens
    /// @param amount of token to stake
    function _stakeInternal(uint256 amount, address interractor) private {
        totalSupplyInternal += amount;
        ++numOfActiveStakesInternal;

        uint256 newId;
        if (interractor != address(this)) {
            newId = ++globalId;
        }
        // else owner stake => newId = 0 (not to affect other stakes)

        stakes[newId] = Stake({
            id: newId,
            amount: amount,
            earned: 0,
            claimed: 0,
            userRewardPerTokenPaid: rewardPerTokenStored,
            owner: interractor,
            timestamp: uint32(block.timestamp),
            unstakedAtBlockNumber: 0,
            unstakedAtBlockTimestamp: 0
        });

        // add new holder if needed (address(this) should not be displayed in holders)
        if (userBalances[interractor] == 0 && interractor != address(this))
            holders.add(interractor);

        userBalances[interractor] += amount;
        idsByUser[interractor].add(newId);
    }

    /// @dev Internal function for withdrawing staked tokens + claim earned rewards
    /// @param id Stake id
    /// @param interractor Address which is supposed to be the stake owner
    /// @return amount - principal token amount, reward - reward token amount
    function _withdrawInternal(
        uint256 id,
        address interractor
    ) private returns (uint256 amount, uint256 reward) {
        Stake storage _stake = stakes[id];
        amount = _stake.amount;
        require(_stake.unstakedAtBlockNumber == 0, 'Already unstaked');
        require(_stake.owner == interractor, 'Called not by stake owner');

        totalSupplyInternal -= amount;
        userBalances[interractor] -= amount;

        --numOfActiveStakesInternal;

        // calculate the available rewards
        // increase the currentRewardPaid
        // earned = 0, increase claimed
        reward = _claimInternal(id, interractor);

        // address(this) should not be displayed in holders
        if (userBalances[interractor] == 0 && interractor != address(this))
            holders.remove(interractor);

        idsByUser[interractor].remove(id);

        if (interractor != address(this)) userInactiveStakes[interractor].push(id);

        stakes[id].unstakedAtBlockNumber = uint32(block.number);
        stakes[id].unstakedAtBlockTimestamp = uint32(block.timestamp);
    }

    function _claimInternal(
        uint256 id,
        address interractor
    ) private returns (uint256 reward) {
        Stake storage _stake = stakes[id];
        require(_stake.unstakedAtBlockNumber == 0, 'Already unstaked');
        require(_stake.owner == interractor, 'Called not by stake owner');

        // reward already updated
        reward = _stake.earned;

        if (reward != 0 && interractor != address(this)) {
            // CLAIM ALL EARNED REWARDS
            // update var only if claim by real user
            currentRewardPaid += reward;
        }

        if (reward != 0) {
            stakes[id].earned = 0;
            stakes[id].claimed += reward;
        }
    }

    /// @dev Update rewards pool information to correct rewards calculation for global varialbes and current stake
    /// @param id stake to update info for
    /// @param updateId flag, true when update stake struct
    function _updateReward(uint256 id, bool updateId) private {
        rewardPerTokenStored = rewardPerToken();
        lastUpdateTime = lastTimeRewardApplicable();

        if (updateId) {
            stakes[id].earned = earned(id);
            stakes[id].userRewardPerTokenPaid = rewardPerTokenStored;
        }
    }
}
